package accountserver

import (
	"encoding/json"
	"errors"
	"fmt"
	"strings"
)

var (
	// ErrUserNotFound is returned when a user object is not found.
	ErrUserNotFound = errors.New("user not found")

	// ErrResourceNotFound is returned when a resource object is not found.
	ErrResourceNotFound = errors.New("resource not found")
)

// It is important to distinguish between different classes of errors,
// so that they can be translated into distinct HTTP status codes and
// transmitted back to the client. Since we also want to retain the
// original error message, we wrap the original error with our custom
// types.

type authError struct {
	error
}

func newAuthError(err error) error {
	return &authError{err}
}

// IsAuthError returns true if err is an authentication /
// authorization error.
func IsAuthError(err error) bool {
	_, ok := err.(*authError)
	return ok
}

type requestError struct {
	error
}

func newRequestError(err error) error {
	return &requestError{err}
}

// IsRequestError returns true if err is a request error (bad
// request).
func IsRequestError(err error) bool {
	_, ok := err.(*requestError)
	return ok
}

type backendError struct {
	error
}

func newBackendError(err error) error {
	return &backendError{err}
}

// IsBackendError returns true if err is a backend error.
func IsBackendError(err error) bool {
	_, ok := err.(*backendError)
	return ok
}

// validationError holds field-specific information that can be
// serialized as JSON if desired.
type validationError struct {
	fields map[string][]string
}

// Initialize or extend a validationError.
func newValidationError(err *validationError, field, msg string) *validationError {
	if err == nil {
		err = &validationError{fields: make(map[string][]string)}
	}
	if field == "" {
		field = "_form"
	}
	err.fields[field] = append(err.fields[field], msg)
	return err
}

func (v *validationError) Error() string {
	var p []string
	if v.fields != nil {
		for f, l := range v.fields {
			p = append(p, fmt.Sprintf("%s: %s", f, strings.Join(l, ", ")))
		}
	} else {
		p = append(p, "INVALID ERROR")
	}
	return fmt.Sprintf("request validation error: %s", strings.Join(p, "; "))
}

func (v *validationError) JSON() []byte {
	data, _ := json.Marshal(v.fields) // nolint
	return data
}

// orNil solves the problem with nil-wrapping interfaces by returning
// an unwrapped nil if the validationError is nil.
func (v *validationError) orNil() error {
	if v == nil {
		return nil
	}
	return v
}
