package accountserver

import (
	"bufio"
	"context"
	"os"
	"strings"
)

// Read non-empty lines from a file, ignoring comments and
// leading/trailing whitespace.
func loadStringListFromFile(path string) ([]string, error) {
	f, err := os.Open(path) // #nosec
	if err != nil {
		return nil, err
	}
	defer f.Close() // nolint: errcheck

	var list []string
	scanner := bufio.NewScanner(f)
	for scanner.Scan() {
		line := scanner.Text()
		if n := strings.Index(line, "#"); n >= 0 {
			line = line[:n]
		}
		line = strings.TrimSpace(line)
		if line != "" {
			list = append(list, line)
		}
	}
	return list, scanner.Err()
}

// A stringSet is just a list of strings with a quick membership test.
type stringSet struct {
	set  map[string]struct{}
	list []string
}

func newStringSetFromList(list []string) *stringSet {
	set := make(map[string]struct{})
	for _, s := range list {
		set[s] = struct{}{}
	}
	return &stringSet{set: set, list: list}
}

func newStringSetFromFileAndList(list []string, path string) (*stringSet, error) {
	if path != "" {
		more, err := loadStringListFromFile(path)
		if err != nil {
			return nil, err
		}
		list = append(list, more...)
	}
	return newStringSetFromList(list), nil
}

func (s stringSet) Contains(needle string) bool {
	_, ok := s.set[needle]
	return ok
}

func (s stringSet) Check(_ *RequestContext, value string) bool {
	_, ok := s.set[value]
	return ok
}

func (s stringSet) List(_ context.Context) []string {
	return s.list
}

// List of email domains backed by our database (with ACLs).
type emailDomainsDataset struct {
	backend Backend
}

func (s *emailDomainsDataset) Check(rctx *RequestContext, value string) bool {
	tx, err := s.backend.NewTransaction()
	if err != nil {
		// Fail close on database errors.
		return false
	}

	// Ignore the error in FindResource() since we're going to fail close anyway.
	// nolint: errcheck
	rsrc, _ := tx.FindResource(rctx.Context, FindResourceRequest{Type: ResourceTypeDomain, Name: value})
	if rsrc == nil {
		return false
	}

	// Only allow domains that have acceptMail=true IF the
	// requesting user is an admin.
	//
	// TODO: relax the ACL check allowing domain owners to create
	// accounts?
	if rsrc.Website.AcceptMail && rctx.Auth.IsAdmin {
		return true
	}

	return false
}

func (s *emailDomainsDataset) List(_ context.Context) []string { return nil }

// Combine multiple datasets.
type multiDataset struct {
	datasets []validationDataset
}

func (m *multiDataset) Check(rctx *RequestContext, value string) bool {
	for _, d := range m.datasets {
		if d.Check(rctx, value) {
			return true
		}
	}
	return false
}

func (m *multiDataset) List(ctx context.Context) []string {
	var out []string
	for _, d := range m.datasets {
		out = append(out, d.List(ctx)...)
	}
	return out
}
