package accountserver

import (
	"context"
	"errors"
	"fmt"
	"regexp"
	"strings"

	"golang.org/x/net/publicsuffix"
)

// Validation and templating require external data providers, lists of
// domains etc, that might be provided by static files or dynamic
// (database-backed) sources, represented by the following interfaces.
type validationDataset interface {
	// Check if a value is contained in the dataset. Since the
	// dataset can be dynamic we pass the full RequestContext
	// (which includes authentication information).
	Check(*RequestContext, string) bool

	// List all values in the dataset. Mostly used for debugging
	// purposes.
	List(context.Context) []string
}

// Some data sources are keyed by resource type, the following helper
// type keeps the code more readable.
type validationDatasetMap map[string]validationDataset

func newValidationDatasetMap(m map[string][]string) validationDatasetMap {
	v := make(validationDatasetMap)
	for key, values := range m {
		v[key] = newStringSetFromList(values)
	}
	return v
}

func (m validationDatasetMap) Check(rctx *RequestContext, key, value string) bool {
	if v, ok := m[key]; ok {
		return v.Check(rctx, value)
	}
	return false
}

func (m validationDatasetMap) List(ctx context.Context, key string) []string {
	if v, ok := m[key]; ok {
		return v.List(ctx)
	}
	return nil
}

// ValidationConfig specifies a large number of validation-related
// configurable parameters.
type ValidationConfig struct {
	ForbiddenUsernames     []string            `yaml:"forbidden_usernames"`
	ForbiddenUsernamesFile string              `yaml:"forbidden_usernames_file"`
	ForbiddenPasswords     []string            `yaml:"forbidden_passwords"`
	ForbiddenPasswordsFile string              `yaml:"forbidden_passwords_file"`
	ForbiddenDomains       []string            `yaml:"forbidden_domains"`
	ForbiddenDomainsFile   string              `yaml:"forbidden_domains_file"`
	AvailableDomains       map[string][]string `yaml:"available_domains"`
	WebsiteRootDir         string              `yaml:"website_root_dir"`
	MinPasswordLen         int                 `yaml:"min_password_len"`
	MaxPasswordLen         int                 `yaml:"max_password_len"`
	MinUsernameLen         int                 `yaml:"min_username_len"`
	MaxUsernameLen         int                 `yaml:"max_username_len"`
	MinUID                 int                 `yaml:"min_backend_uid"`
	MaxUID                 int                 `yaml:"max_backend_uid"`
}

const (
	defaultMinPasswordLen = 8
	defaultMaxPasswordLen = 128
	defaultMinUsernameLen = 3
	defaultMaxUsernameLen = 64
	defaultMinUID         = 1000
	defaultMaxUID         = 0
)

func (c *ValidationConfig) setDefaults() {
	if c.MinPasswordLen == 0 {
		c.MinPasswordLen = defaultMinPasswordLen
	}
	if c.MaxPasswordLen == 0 {
		c.MaxPasswordLen = defaultMaxPasswordLen
	}
	if c.MinUsernameLen == 0 {
		c.MinUsernameLen = defaultMinUsernameLen
	}
	if c.MaxUsernameLen == 0 {
		c.MaxUsernameLen = defaultMaxUsernameLen
	}
	if c.MinUID == 0 {
		c.MinUID = defaultMinUID
	}
	if c.MaxUID == 0 {
		c.MaxUID = defaultMaxUID
	}
}

// The validationContext contains all configuration and backends that
// the various validation functions will need. Most methods on this
// object return functions themselves (ValidatorFunc or variations
// thereof) that can later be called multiple times at will and
// combined with operators like 'allOf'.
type validationContext struct {
	config             *ValidationConfig
	availableDomains   validationDataset
	allowedDomains     validationDataset
	availableShards    validationDataset
	allowedShards      validationDataset
	forbiddenUsernames validationDataset
	forbiddenPasswords validationDataset
	forbiddenDomains   validationDataset
	backend            Backend
}

// ValidatorFunc is the generic interface for unstructured data field
// (string) validators.
type ValidatorFunc func(*RequestContext, string) error

func inSet(set stringSet) ValidatorFunc {
	return func(_ *RequestContext, value string) error {
		if !set.Contains(value) {
			return errors.New("value not allowed")
		}
		return nil
	}
}

func notInSet(set stringSet) ValidatorFunc {
	return func(_ *RequestContext, value string) error {
		if set.Contains(value) {
			return errors.New("invalid value (blacklisted)")
		}
		return nil
	}
}

func minLength(minLen int) ValidatorFunc {
	return func(_ *RequestContext, value string) error {
		if len(value) < minLen {
			return fmt.Errorf("value must be at least %d characters", minLen)
		}
		return nil
	}
}

func maxLength(maxLen int) ValidatorFunc {
	return func(_ *RequestContext, value string) error {
		if len(value) > maxLen {
			return fmt.Errorf("value must be at most %d characters", maxLen)
		}
		return nil
	}
}

func matchRegexp(rx *regexp.Regexp, errmsg string) ValidatorFunc {
	return func(_ *RequestContext, value string) error {
		if !rx.MatchString(value) {
			return errors.New(errmsg)
		}
		return nil
	}
}

// The generic username regexp allows for standard usernames.
var usernameRx = regexp.MustCompile(`^([a-z0-9]+[-_.]?)+[a-z0-9]$`)

func matchUsernameRx() ValidatorFunc {
	return matchRegexp(usernameRx, "value must be an alphanumeric sequence, including . and - characters but not starting or ending with those")
}

func matchSitenameRx() ValidatorFunc {
	return matchUsernameRx()
}

// The identifier regexp is stricter and forbids [-_.] characters.
var identifierRx = regexp.MustCompile(`^[a-z0-9]+$`)

func matchIdentifierRx() ValidatorFunc {
	return matchRegexp(identifierRx, "value must be alphanumeric")
}

// Validator that returns true only if all the arguments return true.
func allOf(funcs ...ValidatorFunc) ValidatorFunc {
	return func(rctx *RequestContext, value string) error {
		for _, f := range funcs {
			if err := f(rctx, value); err != nil {
				return err
			}
		}
		return nil
	}
}

// Separately validate username and domain parts of an email address.
func validateUsernameAndDomain(validateUsername, validateDomain ValidatorFunc) ValidatorFunc {
	return func(rctx *RequestContext, value string) error {
		parts := strings.SplitN(value, "@", 2)
		if len(parts) != 2 {
			return errors.New("malformed email address")
		}
		if err := validateUsername(rctx, parts[0]); err != nil {
			return err
		}
		return validateDomain(rctx, parts[1])
	}
}

// This is not a prescriptive domain name validation regex, it's just used to
// give useful feedback to the user if the domain name looks visibly wrong.
var domainRx = regexp.MustCompile(`^(?:[a-zA-Z0-9_\-]{1,63}\.)+(?:[a-zA-Z]{2,})$`)

func isRegistered(domain string) bool {
	// TODO: ha ha, there isn't really a good way to check.
	return true
}

func validDomainName(_ *RequestContext, value string) error {
	if !domainRx.MatchString(value) {
		return errors.New("invalid domain name")
	}

	// Extract the top-level domain and check that it is registered
	// (which works also if we're hosting the top-level domain
	// ourselves already).
	domain, err := publicsuffix.EffectiveTLDPlusOne(value)
	if err != nil {
		return errors.New("invalid TLD")
	}
	if !isRegistered(domain) {
		return fmt.Errorf("the domain %s does not seem to be registered", domain)
	}
	return nil
}

// Split an email address into username and domain. The address should
// be valid or we will panic.
func splitAddr(addr string) (string, string) {
	parts := strings.SplitN(addr, "@", 2)
	return parts[0], parts[1]
}

// Returns all the possible resources in the email and mailing list
// namespaces that might conflict with the given email address.
func relatedEmails(ctx context.Context, addr string) []FindResourceRequest {
	return []FindResourceRequest{
		{Type: ResourceTypeEmail, Name: addr},
		{Type: ResourceTypeMailingList, Name: addr},
		{Type: ResourceTypeNewsletter, Name: addr},
	}
}

func relatedMailingLists(ctx context.Context, domains validationDatasetMap, addr string) []FindResourceRequest {
	// Check for an email with the same address.
	rel := []FindResourceRequest{
		{Type: ResourceTypeEmail, Name: addr},
	}

	// Mailing lists and newsletters must have unique names
	// regardless of the domain, so we add potential conflicts for
	// mailing lists with the same name over all list-enabled
	// domains.
	user, _ := splitAddr(addr)
	for _, d := range domains.List(ctx, ResourceTypeMailingList) {
		rel = append(rel, FindResourceRequest{
			Type: ResourceTypeMailingList,
			Name: fmt.Sprintf("%s@%s", user, d),
		})
	}
	for _, d := range domains.List(ctx, ResourceTypeNewsletter) {
		rel = append(rel, FindResourceRequest{
			Type: ResourceTypeNewsletter,
			Name: fmt.Sprintf("%s@%s", user, d),
		})
	}
	return rel
}

func relatedWebsites(ctx context.Context, be domainBackend, value string) []FindResourceRequest {
	// Ignore the parent domain (websites share a global namespace).
	return []FindResourceRequest{
		{Type: ResourceTypeWebsite, Name: value},
	}
}

func relatedDomains(ctx context.Context, be domainBackend, value string) []FindResourceRequest {
	return []FindResourceRequest{
		{Type: ResourceTypeDomain, Name: value},
	}
}

func (v *validationContext) isAllowedDomain(rtype string) ValidatorFunc {
	return func(ctx context.Context, value string) error {
		if !v.domains.IsAllowedDomain(ctx, rtype, value) {
			return errors.New("unavailable domain")
		}
		return nil
	}
}

func isResourceAvailable(ctx context.Context, backend Backend, rsrcs []FindResourceRequest) bool {
	tx, err := backend.NewTransaction()
	if err != nil {
		// Fail close, return "not available".
		return false
	}
	ok, _ := tx.HasAnyResource(ctx, rsrcs)
	return !ok
}

func (v *validationContext) isAvailableDAV() ValidatorFunc {
	return func(ctx context.Context, value string) error {
		rel := []FindResourceRequest{
			{
				Type: ResourceTypeDAV,
				Name: value,
			},
		}

		// Run the presence check in a new transaction. Unavailability
		// of the server results in a validation error (fail close).
		tx, err := v.backend.NewTransaction()
		if err != nil {
			return err
		}
		// Errors will cause to consider the resource unavailable.
		if ok, _ := tx.HasAnyResource(ctx, rel); ok { // nolint
			return errors.New("name unavailable")
		}
		return nil
	}
}

func (v *validationContext) isAvailableDatabase() ValidatorFunc {
	return func(ctx context.Context, value string) error {
		rel := []FindResourceRequest{
			{
				Type: ResourceTypeDatabase,
				Name: value,
			},
		}

		// Run the presence check in a new transaction. Unavailability
		// of the server results in a validation error (fail close).
		tx, err := v.backend.NewTransaction()
		if err != nil {
			return err
		}
		// Errors will cause to consider the resource unavailable.
		if ok, _ := tx.HasAnyResource(ctx, rel); ok { // nolint
			return errors.New("name unavailable")
		}
		return nil
	}
}

func (v *validationContext) validHostedEmail() ValidatorFunc {
	return validateUsernameAndDomain(
		allOf(
			matchUsernameRx(),
			minLength(v.config.MinUsernameLen),
			maxLength(v.config.MaxUsernameLen),
			notInSet(v.config.forbiddenUsernames),
		),
		allOf(v.isAllowedDomain(ResourceTypeEmail)),
	)
}

func (v *validationContext) validHostedNewEmail() ValidatorFunc {
	return allOf(
		v.validHostedEmail(),
		v.isAvailableEmailAddr(),
	)
}

func (v *validationContext) validHostedMailingList() ValidatorFunc {
	return validateUsernameAndDomain(
		allOf(
			matchUsernameRx(),
			minLength(v.config.MinUsernameLen),
			maxLength(v.config.MaxUsernameLen),
			notInSet(v.config.forbiddenUsernames),
		),
		allOf(v.isAllowedDomain(ResourceTypeMailingList)),
	)
}

func (v *validationContext) validPassword() ValidatorFunc {
	return allOf(
		minLength(v.config.MinPasswordLen),
		maxLength(v.config.MaxPasswordLen),
		notInSet(v.config.forbiddenPasswords),
	)
}

// ResourceValidator is a composite type validator that checks
// various fields in a Resource, depending on its type.
//
// The availability check is separate and it performs the role of an
// ACL check.
type resourceValidator interface {
	Validate(*RequestContext, *Resource, *User) error
	IsAvailable(*RequestContext, *Resource, *User) bool
}

type resourceValidatorBase struct {
	backend Backend
}

// Validate the common attributes of a Resource.
func (v *resourceValidatorBase) Validate(_ *RequestContext, r *Resource, user *User) error {
	// Validate the status enum.
	switch r.Status {
	case ResourceStatusActive, ResourceStatusInactive, ResourceStatusReadonly, ResourceStatusArchived:
	default:
		return errors.New("unknown resource status")
	}

	// If the resource has a ParentID, it must reference another
	// resource owned by the user.
	if !r.ParentID.Empty() {
		if user == nil {
			return errors.New("resource can't have parent without user context")
		}
		if p := user.GetResourceByID(r.ParentID); p == nil {
			return errors.New("parent references unknown resource")
		}
	}

	return nil
}

// Helper function to check availability of one or more resources in
// the database. We accept a list of resources as input because the
// name spaces for different resources partially overlap (emails,
// mailing lists, etc) so we often have to check multiple types.
func (v *resourceValidatorBase) checkResources(ctx context.Context, rsrcs []FindResourcesRequest) bool {
	tx, err := v.backend.NewTransaction()
	if err != nil {
		// Fail close on database errors.
		return false
	}
	ok, _ := tx.HasAnyResource(ctx, rsrcs)
	return ok
}

type shardedResourceValidatorBase struct {
	*resourceValidatorBase
	allowedShards validationDatasetMap
}

// Validate the attributes of a sharded resource.
func (v *shardedResourceValidatorBase) Validate(rctx *ResourceContext, r *Resource, user *User) error {
	if err := v.resourceValidatorBase.Validate(rctx, r, user); err != nil {
		return err
	}

	if r.Shard == "" {
		return errors.New("empty shard")
	}
	if !allowedShards.Check(rctx, r.Type, r.Shard) {
		return fmt.Errorf(
			"invalid shard %s for resource type %s (allowed: %v)",
			r.Shard,
			r.Type,
			allowedShards.List(rctx.Context(), r.Type),
		)
	}
	if r.OriginalShard == "" {
		return errors.New("empty original_shard")
	}
	return nil
}

// Validator for Email resources.
type emailResourceValidator struct {
	*shardedResourceValidatorBase

	maxUsernameLen     int
	minUsernameLen     int
	forbiddenUsernames validationDataset
	allowedDomains     validationDataset
}

func (v *emailResourceValidator) Validate(rctx *ResourceContext, r *Resource, user *User) error {
	if err := v.shardedResourceValidatorBase.Validate(rctx, r, user); err != nil {
		return err
	}

	// Email resources shouldn't be nested.
	if !r.ParentID.Empty() {
		return errors.New("resource should not have parent")
	}

	if r.Email == nil {
		return errors.New("resource has no email metadata")
	}
	if r.Email.Maildir == "" {
		return errors.New("empty maildir")
	}

	// Validate the resource email address field, and check for
	// availability of the domain.
	return validateUsernameAndDomain(
		allOf(
			matchUsernameRx(),
			minLength(v.minUsernameLen),
			maxLength(v.maxUsernameLen),
			notInSet(v.forbiddenUsernames),
		),
		allOf(
			inSet(v.allowedDomains),
		),
	)(rtcx, r.Name)
}

func (v *emailResourceValidator) IsAvailable(rctx *ResourceContext, r *Resource, user *User) bool {
	// Emails share the same namespace as mailing lists and newsletters.
	addr := r.Name
	return v.checkResources([]FindResourceRequest{
		{Type: ResourceTypeEmail, Name: addr},
		{Type: ResourceTypeMailingList, Name: addr},
		{Type: ResourceTypeNewsletter, Name: addr},
	})
}

// Mailing lists and newsletters share the same validation logic.
type listResourceValidatorBase struct {
	*shardedResourceValidatorBase
}

func (v *listResourceValidatorBase) Validate(rctx *RequestContext, r *Resource, user *User) error {
	if err := v.shardedResourceValidatorBase.Validate(rctx, r, user); err != nil {
		return err
	}

	// List resources shouldn't be nested.
	if !r.ParentID.Empty() {
		return errors.New("resource should not have parent")
	}

	// Validate the list name.
	return validateUsernameAndDomain(
		allOf(
			matchUsernameRx(),
			minLength(v.minUsernameLen),
			maxLength(v.maxUsernameLen),
			notInSet(v.forbiddenUsernames),
		),
		inSet(v.allowedDomains),
	)(rctx, r.Name)
}

// Helper to validate a list of admins.
func (v *listResourceValidatorBase) validateAdmins(rctx *RequestContext, admins []string) error {
	if len(admins) < 1 {
		return errors.New("list has no admins")
	}
	for _, addr := range admins {
		if err := validateEmailAddr(rctx, addr); err != nil {
			return err
		}
	}
	return nil
}

func (v *listResourceValidatorBase) IsAvailable(rctx *ResourceContext, r *Resource, _ *User) bool {
	// Check for an email with the same address.
	rel := []FindResourceRequest{
		{Type: ResourceTypeEmail, Name: r.Name},
	}

	// Mailing lists and newsletters must have unique names
	// regardless of the domain, so we add potential conflicts for
	// mailing lists with the same name over all list-enabled
	// domains.
	user, _ := splitAddr(r.Name)
	for _, d := range domains.List(ctx, ResourceTypeMailingList) {
		rel = append(rel, FindResourceRequest{
			Type: ResourceTypeMailingList,
			Name: fmt.Sprintf("%s@%s", user, d),
		})
	}
	for _, d := range domains.List(ctx, ResourceTypeNewsletter) {
		rel = append(rel, FindResourceRequest{
			Type: ResourceTypeNewsletter,
			Name: fmt.Sprintf("%s@%s", user, d),
		})
	}
	return v.checkResources(rel)
}

// Validator for MailingList resources.
type mailingListResourceValidator struct {
	*listResourceValidatorBase
}

func (v *mailingListResourceValidator) Validate(rctx *RequestContext, r *Resource, user *User) error {
	if err := v.listResourceValidatorBase.Validate(rctx, r, user); err != nil {
		return err
	}

	if r.List == nil {
		return errors.New("resource has no list metadata")
	}
	return v.validateAdmins(rctx, r.List.Admins)
}

// Validator for Newsletter resources.
type newsletterResourceValidator struct {
	*listResourceValidatorBase
}

func (v *newsletterResourceValidator) Validate(rctx *RequestContext, r *Resource, user *User) error {
	if err := v.listResourceValidatorBase.Validate(rctx, r, user); err != nil {
		return err
	}

	if r.Newsletter == nil {
		return errors.New("resource has no newsletter metadata")
	}
	return r.validateAdmins(rctx, r.Newsletter.Admins)
}

// Validator helper type for resources that have a UNIX user ID.
type uidResourceValidatorMixin struct {
	minUID, maxUID int
}

func (v *uidResourceValidatorMixin) validateUID(uid int, user *User) error {
	if uid == 0 {
		return errors.New("resource UID is not set")
	}
	if uid < v.minUID || (v.maxUID > 0 && uid > v.maxUID) {
		return fmt.Errorf("resource UID %d outside of allowed range (%d-%d)", uid, v.minUID, v.maxUID)
	}
	if user != nil && uid != user.UID {
		return errors.New("resource UID differs from user UID")
	}
	return nil
}

// Base validator type for web resources.
type webResourceValidatorBase struct {
	*shardedResourceValidatorBase
	*uidResourceValidatorMixin

	webroot string
}

func (v *webResourceValidatorBase) Validate(rctx *RequestContext, r *Resource, user *User) error {
	if err := v.shardedResourceValidatorBase.Validate(rctx, r, user); err != nil {
		return err
	}

	// Web resources shouldn't be nested.
	if !r.ParentID.Empty() {
		return errors.New("resource should not have parent")
	}

	if r.Website == nil {
		return errors.New("resource has no website metadata")
	}

	// Document root checks: must not be empty, and the
	// user must have a DAV account with a home directory
	// that is a parent of this document root.
	if r.Website.DocumentRoot == "" {
		return errors.New("empty document_root")
	}
	if !isSubdir(v.webroot, r.Website.DocumentRoot) {
		return errors.New("document root outside of web root")
	}
	if !hasMatchingDAVAccount(user, r) {
		return errors.New("website has no matching DAV account")
	}

	return v.uidResourceValidatorMixin.validateUID(r.Website.UID, user)
}

type websiteResourceValidator struct {
	*webResourceValidatorBase

	allowedDomains     validatorDataset
	forbiddenSiteNames validatorDataset
	minNameLen         int
}

func (v *websiteResourceValidator) Validate(rctx *RequestContext, r *Resource, user *User) error {
	if err := v.webResourceValidatorBase.Validate(rctx, r, user); err != nil {
		return err
	}

	if err := inSet(v.allowedDomains)(rctx, r.Website.ParentDomain); err != nil {
		return err
	}
	if err := allOf(
		matchSitenameRx(),
		minLength(v.minNameLen),
		notInSet(v.forbiddenSiteNames),
	)(rctx, r.Name); err != nil {
		return err
	}

	return nil
}

func (v *websiteResourceValidatorBase) IsAvailable(rctx *ResourceContext, r *Resource, _ *User) bool {
}

type domainResourceValidator struct {
	*webResourceValidatorBase

	forbiddenDomains validationDataset
}

func (v *domainResourceValidator) Validate(rctx *RequestContext, r *Resource, user *User) error {
	if err := v.webResourceValidatorBase.Validate(rctx, r, user); err != nil {
		return err
	}

	if r.Website.ParentDomain != "" {
		return errors.New("non-empty parent_domain on domain resource")
	}

	if err := allOf(
		matchDomainRx(),
		notInSet(v.forbiddenDomains),
	)(rctx, r.Name); err != nil {
		return err
	}

	return nil
}

func (v *domainResourceValidatorBase) IsAvailable(rctx *ResourceContext, r *Resource, _ *User) bool {
}

type davResourceValidator struct {
	*shardedResourceValidatorBase
	*uidResourceValidatorMixin

	minLen  int
	webroot string
}

func (v *davResourceValidator) Validate(rctx *RequestContext, r *Resource, user *User) error {
	// return firstErr(
	// 	v.shardedResourceValidatorBase.Validate(rctx, r, user),
	// 	noParentID,
	// 	notNil(r.DAV),
	// 	validateField(r.Name, allOf(
	// 		minLength(v.minLen),
	// 		matchIdentifierRx(),
	// 	)),
	// 	subdirOf(v.webroot, r.DAV.Homedir),
	// )

	if err := v.shardedResourceValidatorBase.Validate(rctx, r, user); err != nil {
		return err
	}

	// DAV resources shouldn't be nested.
	if !r.ParentID.Empty() {
		return errors.New("resource should not have parent")
	}

	if r.DAV == nil {
		return errors.New("resource has no DAV metadata")
	}

	if err := allOf(
		minLength(v.minLen),
		matchIdentifierRx(),
	)(rctx, r.Name); err != nil {
		return err
	}
	if !isSubdir(v.webroot, r.DAV.Homedir) {
		return errors.New("homedir outside of web root")
	}

	return v.uidResourceValidatorMixin.validateUID(r.DAV.UID, user)
}

///////////////////// CUT THE CRAP HERE /////////////////////////

func findMatchingDAVAccount(user *User, r *Resource) *Resource {
	for _, dav := range user.GetResourcesByType(ResourceTypeDAV) {
		if isSubdir(dav.DAV.Homedir, r.Website.DocumentRoot) {
			return dav
		}
	}
	return nil
}

func hasMatchingDAVAccount(user *User, r *Resource) bool {
	return findMatchingDAVAccount(user, r) != nil
}

func (v *validationContext) validateUID(uid int, user *User) error {
	if uid == 0 {
		return errors.New("uid is not set")
	}
	if uid < v.config.MinUID || (v.config.MaxUID > 0 && uid > v.config.MaxUID) {
		return fmt.Errorf("uid %d outside of allowed range (%d-%d)", uid, v.config.MinUID, v.config.MaxUID)
	}
	if user != nil && uid != user.UID {
		return errors.New("uid of resource differs from uid of user")
	}
	return nil
}

func (v *validationContext) validateWebsiteCommon(r *Resource, user *User) error {
	// Document root checks: must not be empty, and the
	// user must have a DAV account with a home directory
	// that is a parent of this document root.
	if r.Website.DocumentRoot == "" {
		return errors.New("empty document_root")
	}
	if !isSubdir(v.config.WebsiteRootDir, r.Website.DocumentRoot) {
		return errors.New("document root outside of web root")
	}
	if !hasMatchingDAVAccount(user, r) {
		return errors.New("website has no matching DAV account")
	}

	// UID checks.
	return v.validateUID(r.Website.UID, user)
}

func (v *validationContext) validDomainResource() ResourceValidatorFunc {
	domainValidator := allOf(
		minLength(6),
		validDomainName,
		notInSet(v.config.forbiddenDomains),
	)
	newDomainValidator := allOf(
		domainValidator,
		v.isAvailableDomain(),
	)

	return func(ctx context.Context, r *Resource, user *User, isNew bool) error {
		if err := v.validateShardedResource(ctx, r, user); err != nil {
			return err
		}

		// Web resources aren't nested.
		if !r.ParentID.Empty() {
			return errors.New("resource should not have parent")
		}

		if r.Website == nil {
			return errors.New("resource has no website metadata")
		}

		var err error
		if isNew {
			err = newDomainValidator(ctx, r.Name)
		} else {
			err = domainValidator(ctx, r.Name)
		}
		if err != nil {
			return err
		}

		if r.Website.ParentDomain != "" {
			return errors.New("non-empty parent_domain on domain resource")
		}

		return v.validateWebsiteCommon(r, user)
	}
}

func (v *validationContext) validWebsiteResource() ResourceValidatorFunc {
	nameValidator := allOf(
		minLength(6),
		matchSitenameRx(),
	)
	newNameValidator := allOf(
		nameValidator,
		v.isAvailableWebsite(),
	)
	parentValidator := v.isAllowedDomain(ResourceTypeWebsite)

	return func(ctx context.Context, r *Resource, user *User, isNew bool) error {
		if err := v.validateShardedResource(ctx, r, user); err != nil {
			return err
		}

		// Web resources aren't nested.
		if !r.ParentID.Empty() {
			return errors.New("resource should not have parent")
		}

		if r.Website == nil {
			return errors.New("resource has no website metadata")
		}

		var err error
		if isNew {
			err = newNameValidator(ctx, r.Name)
		} else {
			err = nameValidator(ctx, r.Name)
		}
		if err != nil {
			return err
		}
		if err := parentValidator(ctx, r.Website.ParentDomain); err != nil {
			return err
		}

		return v.validateWebsiteCommon(r, user)
	}
}

func (v *validationContext) validDAVResource() ResourceValidatorFunc {
	davValidator := allOf(
		minLength(4),
		matchIdentifierRx(),
	)
	newDAVValidator := allOf(
		davValidator,
		v.isAvailableDAV(),
	)
	return func(ctx context.Context, r *Resource, user *User, isNew bool) error {
		if err := v.validateShardedResource(ctx, r, user); err != nil {
			return err
		}

		// DAV resources aren't nested.
		if !r.ParentID.Empty() {
			return errors.New("resource should not have parent")
		}

		var err error
		if isNew {
			err = newDAVValidator(ctx, r.Name)
		} else {
			err = davValidator(ctx, r.Name)
		}
		if err != nil {
			return err
		}

		if r.DAV == nil {
			return errors.New("resource has no dav metadata")
		}
		if !isSubdir(v.config.WebsiteRootDir, r.DAV.Homedir) {
			return errors.New("homedir outside of web root")
		}

		return v.validateUID(r.DAV.UID, user)
	}
}

func (v *validationContext) validDatabaseResource() ResourceValidatorFunc {
	dbValidator := allOf(
		minLength(4),
		matchIdentifierRx(),
	)
	newDBValidator := allOf(
		dbValidator,
		v.isAvailableDatabase(),
	)
	return func(ctx context.Context, r *Resource, user *User, isNew bool) error {
		if err := v.validateShardedResource(ctx, r, user); err != nil {
			return err
		}

		// Database resources must be nested below a website.
		if r.ParentID.Empty() {
			return errors.New("database resources should be nested")
		}

		// TODO: move to global validation
		// switch r.ParentID.Type() {
		// case ResourceTypeWebsite, ResourceTypeDomain:
		// default:
		// 	return errors.New("database parent is not a website resource")
		// }

		var err error
		if isNew {
			err = newDBValidator(ctx, r.Name)
		} else {
			err = dbValidator(ctx, r.Name)
		}
		if err != nil {
			return err
		}

		if r.Database == nil {
			return errors.New("resource has no database metadata")
		}
		return nil
	}
}

// Validator for arbitrary resource types.
type resourceValidator struct {
	rvs map[string]ResourceValidatorFunc
}

func newResourceValidator(v *validationContext) *resourceValidator {
	return &resourceValidator{
		rvs: map[string]ResourceValidatorFunc{
			ResourceTypeEmail:       v.validEmailResource(),
			ResourceTypeMailingList: v.validListResource(),
			ResourceTypeNewsletter:  v.validNewsletterResource(),
			ResourceTypeDomain:      v.validDomainResource(),
			ResourceTypeWebsite:     v.validWebsiteResource(),
			ResourceTypeDAV:         v.validDAVResource(),
			ResourceTypeDatabase:    v.validDatabaseResource(),
		},
	}
}

func (v *resourceValidator) validateResource(ctx context.Context, r *Resource, user *User, isNew bool) error {
	// Obvious basic sanity checks on the Resource parameters.
	if r.Name == "" {
		return errors.New("resource name unset")
	}
	if r.Type == "" {
		return errors.New("resource type unset")
	}

	rv, ok := v.rvs[r.Type]
	if !ok {
		return fmt.Errorf("unknown resource type '%s'", r.Type)
	}

	return rv(ctx, r, user, isNew)
}

// Common validators for specific field types.
type fieldValidators struct {
	password ValidatorFunc
	newEmail ValidatorFunc
}

func newFieldValidators(v *validationContext) *fieldValidators {
	return &fieldValidators{
		password: v.validPassword(),
		newEmail: v.validHostedNewEmail(),
	}
}

// UserValidatorFunc is a compound validator for User objects.
type UserValidatorFunc func(context.Context, *User, bool) error

// Verify that user-level invariants are respected. This check can be applied
// to new or existing objects.
//
// nolint: gocyclo
func checkUserInvariants(user *User) error {
	// Count email resources. An account must have exactly 1 email
	// resource.
	var email *Resource
	var emailCount int
	for _, rsrc := range user.Resources {
		if rsrc.Type == ResourceTypeEmail {
			email = rsrc
			emailCount++
		}
	}
	if emailCount == 0 {
		return errors.New("account is missing email resource")
	}
	if emailCount > 1 {
		return errors.New("account can't have more than one email resource")
	}
	if email.Name != user.Name {
		return errors.New("email and username do not match")
	}

	// Check UID.
	if user.UID == 0 {
		return errors.New("UID unset")
	}
	// Check that all UIDs are the same.
	for _, rsrc := range user.Resources {
		switch rsrc.Type {
		case ResourceTypeWebsite, ResourceTypeDomain:
			if rsrc.Website.UID != user.UID {
				return fmt.Errorf("UID of resource %s does not match user (%d vs %d)", rsrc.String(), rsrc.Website.UID, user.UID)
			}
		case ResourceTypeDAV:
			if rsrc.DAV.UID != user.UID {
				return fmt.Errorf("UID of resource %s does not match user (%d vs %d)", rsrc.String(), rsrc.DAV.UID, user.UID)
			}
		}
	}

	return nil
}

// A custom validator for new User objects.
func (v *validationContext) validUser() UserValidatorFunc {
	nameValidator := v.validHostedEmail()
	newNameValidator := v.validHostedNewEmail()
	return func(ctx context.Context, user *User, isNew bool) error {
		var err error
		if isNew {
			err = newNameValidator(ctx, user.Name)
		} else {
			err = nameValidator(ctx, user.Name)
		}
		if err != nil {
			return err
		}
		return checkUserInvariants(user)
	}
}

func isSubdir(root, dir string) bool {
	return strings.HasPrefix(dir, root+"/")
}
