package server

import (
	"fmt"
	"log"
	"strings"
	"sync"
	"time"

	"git.autistici.org/id/auth"
	"git.autistici.org/id/auth/backend"
	"github.com/prometheus/client_golang/prometheus"
)

// Try to use as little memory as possible for each entry: use a UNIX
// timestamp instead of a time.Time, and use an int32 as a saturating
// counter.
type ratelimitDatum struct {
	stamp   int64
	counter int32
}

func (d ratelimitDatum) age(now int64) int64 {
	return now - d.stamp
}

// Ratelimiter is a simple counter-based rate limiter, allowing the
// first N requests over each period of time T.
type Ratelimiter struct {
	limit  int32
	period int64

	mx sync.Mutex
	c  map[string]ratelimitDatum
}

func newRatelimiter(limit, period int) *Ratelimiter {
	r := &Ratelimiter{
		limit:  int32(limit),
		period: int64(period),
		c:      make(map[string]ratelimitDatum),
	}
	go r.expungeThread()
	return r
}

// AllowIncr performs a check and an increment at the same time, while
// holding a mutex, so it is robust in face of concurrent requests.
func (r *Ratelimiter) AllowIncr(key string) bool {
	if key == "" {
		return true
	}
	r.mx.Lock()
	d := r.get(key)
	var allowed bool
	if d.counter <= r.limit {
		allowed = true
		d.counter++
		r.set(key, d)
	}
	r.mx.Unlock()

	return allowed
}

func (r *Ratelimiter) get(key string) ratelimitDatum {
	now := time.Now().Unix()
	d, ok := r.c[key]
	if !ok || d.age(now) > r.period {
		d = ratelimitDatum{stamp: now}
		r.c[key] = d
	}
	return d
}

func (r *Ratelimiter) set(key string, d ratelimitDatum) {
	r.c[key] = d
}

func (r *Ratelimiter) expunge() {
	cutoff := time.Now().Unix() - 2*r.period
	r.mx.Lock()
	for k, d := range r.c {
		if d.stamp < cutoff {
			delete(r.c, k)
		}
	}
	r.mx.Unlock()
}

var ratelimitExpungePeriod = 300 * time.Second

func (r *Ratelimiter) expungeThread() {
	for range time.NewTicker(ratelimitExpungePeriod).C {
		r.expunge()
	}
}

// Blacklist can blacklist keys whose request rate is above a
// specified threshold.
type Blacklist struct {
	name   string
	r      *Ratelimiter
	bl     map[string]int64
	blTime int64
}

func newBlacklist(name string, limit, period, blacklistTime int) *Blacklist {
	return &Blacklist{
		r:      newRatelimiter(limit, period),
		bl:     make(map[string]int64),
		blTime: int64(blacklistTime),
	}
}

// Allow returns true if this request (identified by the given key)
// should be allowed.
func (b *Blacklist) Allow(key string) bool {
	if key == "" {
		return true
	}

	b.r.mx.Lock()
	deadline, ok := b.bl[key]
	if ok && deadline < time.Now().Unix() {
		delete(b.bl, key)
		ok = false
	}
	b.r.mx.Unlock()
	return !ok
}

// Incr increments the counter for the given key for the current time
// period.
func (b *Blacklist) Incr(key string) {
	if key == "" {
		return
	}

	// Count one higher than limit, and trigger the blacklist when
	// we reach that.
	limitp1 := b.r.limit + 1

	b.r.mx.Lock()
	d := b.r.get(key)
	if d.counter < limitp1 {
		d.counter++
		b.r.set(key, d)
	} else if d.counter == limitp1 {
		log.Printf("blacklisted %s", key)
		blacklistCounter.WithLabelValues(b.name).Inc()
		b.bl[key] = time.Now().Unix() + b.blTime
	}
	b.r.mx.Unlock()
}

// Function that extracts a key from a request.
type ratelimitKeyFunc func(*backend.User, *auth.Request) string

// Extract the username from the request.
func usernameKey(user *backend.User, _ *auth.Request) string {
	return user.Name
}

// Extract the client IP address (if present) from the request.
func ipAddrKey(_ *backend.User, req *auth.Request) string {
	if req.DeviceInfo != nil {
		return req.DeviceInfo.RemoteAddr
	}
	return ""
}

func parseKeyFunc(k string) (ratelimitKeyFunc, error) {
	switch k {
	case "ip":
		return ipAddrKey, nil
	case "user":
		return usernameKey, nil
	default:
		return nil, fmt.Errorf("unknown key %s", k)
	}
}

// Configuration for a rate limiter or blacklist (depending on whether
// BlacklistTime is set or not). All times are specified in seconds.
type authRatelimiterConfig struct {
	Limit         int      `yaml:"limit"`
	Period        int      `yaml:"period"`
	BlacklistTime int      `yaml:"blacklist_for"`
	OnFailure     bool     `yaml:"on_failure"`
	Keys          []string `yaml:"keys"`
	Bypass        []struct {
		Key   string `yaml:"key"`
		Value string `yaml:"value"`
	} `yaml:"bypass"`
}

type bypassEvaluator struct {
	keyFn ratelimitKeyFunc
	value string
}

func (r *authRatelimiterConfig) bypassFunctions() ([]bypassEvaluator, error) {
	out := make([]bypassEvaluator, 0, len(r.Bypass))
	for _, b := range r.Bypass {
		f, err := parseKeyFunc(b.Key)
		if err != nil {
			return nil, err
		}
		out = append(out, bypassEvaluator{
			keyFn: f,
			value: b.Value,
		})
	}
	return out, nil
}

func (r *authRatelimiterConfig) keyFunctions() ([]ratelimitKeyFunc, error) {
	var keyFuncs []ratelimitKeyFunc
	for _, k := range r.Keys {
		f, err := parseKeyFunc(k)
		if err != nil {
			return nil, err
		}
		keyFuncs = append(keyFuncs, f)
	}
	return keyFuncs, nil
}

const rlKeySep = ";"

type authRatelimiterBase struct {
	keyFuncs []ratelimitKeyFunc
	bypass   []bypassEvaluator
}

func newAuthRatelimiterBase(config *authRatelimiterConfig) (*authRatelimiterBase, error) {
	kf, err := config.keyFunctions()
	if err != nil {
		return nil, err
	}
	be, err := config.bypassFunctions()
	if err != nil {
		return nil, err
	}
	return &authRatelimiterBase{
		keyFuncs: kf,
		bypass:   be,
	}, nil
}

func (r *authRatelimiterBase) shouldBypass(user *backend.User, req *auth.Request) bool {
	for _, b := range r.bypass {
		if b.value == b.keyFn(user, req) {
			return true
		}
	}
	return false
}

func (r *authRatelimiterBase) key(user *backend.User, req *auth.Request) string {
	if len(r.keyFuncs) == 1 {
		return r.keyFuncs[0](user, req)
	}

	var parts []string
	for _, f := range r.keyFuncs {
		parts = append(parts, f(user, req))
	}
	return strings.Join(parts, rlKeySep)
}

// Request-oriented ratelimiter with configurable keys.
type authRatelimiter struct {
	*authRatelimiterBase
	rl *Ratelimiter
}

func newAuthRatelimiter(config *authRatelimiterConfig) (*authRatelimiter, error) {
	r, err := newAuthRatelimiterBase(config)
	if err != nil {
		return nil, err
	}
	return &authRatelimiter{
		authRatelimiterBase: r,
		rl:                  newRatelimiter(config.Limit, config.Period),
	}, nil
}

func (r *authRatelimiter) AllowIncr(user *backend.User, req *auth.Request) bool {
	if r.shouldBypass(user, req) {
		return true
	}
	key := r.key(user, req)
	if key == "" { // An empty key bypasses the rate limit.
		return true
	}
	return r.rl.AllowIncr(key)
}

// Request-oriented blacklist with configurable keys.
type authBlacklist struct {
	*authRatelimiterBase
	bl        *Blacklist
	onFailure bool
}

func newAuthBlacklist(name string, config *authRatelimiterConfig) (*authBlacklist, error) {
	r, err := newAuthRatelimiterBase(config)
	if err != nil {
		return nil, err
	}
	return &authBlacklist{
		authRatelimiterBase: r,
		bl:                  newBlacklist(name, config.Limit, config.Period, config.BlacklistTime),
		onFailure:           config.OnFailure,
	}, nil
}

func (b *authBlacklist) Allow(user *backend.User, req *auth.Request) bool {
	if b.shouldBypass(user, req) {
		return true
	}
	key := b.key(user, req)
	if key == "" { // An empty key bypasses the blacklist.
		return true
	}
	return b.bl.Allow(key)
}

func (b *authBlacklist) Incr(user *backend.User, req *auth.Request, resp *auth.Response) {
	if b.onFailure && resp != nil && resp.Status == auth.StatusOK {
		return
	}
	b.bl.Incr(b.key(user, req))
}

var blacklistCounter = prometheus.NewCounterVec(prometheus.CounterOpts{
	Name: "auth_blacklisted_total",
	Help: "Counter of blacklisted entries.",
}, []string{"bl"})

func init() {
	prometheus.MustRegister(blacklistCounter)
}
